import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
import cv2
from scipy.spatial.distance import cosine
from typing import Union, Tuple, Literal
import os
from .preprocessing import preprocess_image, get_face_landmarks, align_face
from .architectures import LightCNN_V4, LightCNN_29Layers_v2, resblock
from .utils import download_weights

ModelType = Literal["LightCNN-V4", "LightCNN_29Layers_V2"]

class LightCNN:
    def __init__(self, model_name: ModelType = "LightCNN-V4", device: str = None):
        self.model_name = model_name
        if device is None:
            self.device = "cuda" if torch.cuda.is_available() else "cpu"
        else:
            self.device = device
            
        download_weights()
        self.model = self._load_model()
        self.model.eval()
        self.model.to(self.device)

    def _load_model(self):
        if self.model_name == "LightCNN-V4":
            model = LightCNN_V4()
            weights_file = "LightCNN-V4_checkpoint.pth.tar"
        else:  # LightCNN_29Layers_V2
            model = LightCNN_29Layers_v2(num_classes=80013)  # Specific num_classes for pretrained model
            weights_file = "LightCNN_29Layers_V2_checkpoint.pth.tar"

        weights_path = os.path.join(
            os.path.dirname(__file__), 
            "weights", 
            weights_file
        )

        try:
            checkpoint = torch.load(weights_path, map_location=self.device)
            if "state_dict" in checkpoint:
                # Clean state dict keys
                new_state_dict = {}
                for k, v in checkpoint["state_dict"].items():
                    key = k.replace("module.", "")
                    # Only load feature extraction layers
                    if "fc2" not in key and "fc" not in key:
                        new_state_dict[key] = v
                model.load_state_dict(new_state_dict, strict=False)
            else:
                raise ValueError("Invalid checkpoint format")
        except Exception as e:
            print(f"Error loading weights: {e}")
            raise

        model.eval()  # Set to evaluation mode
        return model

    def get_features(self, image: Union[str, np.ndarray], align: bool = True) -> np.ndarray:
        """Extract features from a face image"""
        if isinstance(image, str):
            image = cv2.imread(image)
            if image is None:
                raise ValueError("Could not load image")

        if align:
            # Convert to grayscale for face detection
            gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
            # Detect faces
            faces = self.detector(gray)
            if len(faces) == 0:
                raise ValueError("No face detected in image")
            
            # Get facial landmarks
            landmarks = self.predictor(gray, faces[0])
            
            # Align face using landmarks
            left_eye = (landmarks.part(36).x, landmarks.part(36).y)
            right_eye = (landmarks.part(45).x, landmarks.part(45).y)
            
            # Calculate angle for alignment
            dx = right_eye[0] - left_eye[0]
            dy = right_eye[1] - left_eye[1]
            angle = np.degrees(np.arctan2(dy, dx))
            
            # Get center point between eyes
            center = ((left_eye[0] + right_eye[0]) // 2,
                     (left_eye[1] + right_eye[1]) // 2)
            
            # Get rotation matrix
            M = cv2.getRotationMatrix2D(center, angle, 1.0)
            
            # Perform the rotation
            aligned_image = cv2.warpAffine(image, M, (image.shape[1], image.shape[0]))
            
            # Crop the face region using landmarks
            x_min = max(0, min(landmarks.part(j).x for j in range(68)))
            y_min = max(0, min(landmarks.part(j).y for j in range(68)))
            x_max = min(image.shape[1], max(landmarks.part(j).x for j in range(68)))
            y_max = min(image.shape[0], max(landmarks.part(j).y for j in range(68)))
            
            # Add some margin
            margin = int((x_max - x_min) * 0.1)  # 10% margin
            x_min = max(0, x_min - margin)
            y_min = max(0, y_min - margin)
            x_max = min(image.shape[1], x_max + margin)
            y_max = min(image.shape[0], y_max + margin)
            
            image = aligned_image[y_min:y_max, x_min:x_max]

        img_tensor = preprocess_image(image, self.model_name)
        img_tensor = img_tensor.to(self.device)

        with torch.no_grad():
            self.model.eval()
            # Get features from the model
            if self.model_name == "LightCNN_29Layers_V2":
                features = self.model(img_tensor)[0]  # Get first element of tuple
            else:
                features = self.model(img_tensor)
            
            # Normalize features
            features = F.normalize(features, p=2, dim=1)

        return features.cpu().numpy().flatten()

    def verify(
        self,
        image1: Union[str, np.ndarray],
        image2: Union[str, np.ndarray],
        align: bool = True,
    ) -> Tuple[float, bool]:
        """Verify if two face images belong to the same person"""
        feat1 = self.get_features(image1, align=align)
        feat2 = self.get_features(image2, align=align)

        # Normalize features
        feat1 = feat1 / np.linalg.norm(feat1)
        feat2 = feat2 / np.linalg.norm(feat2)
        
        # Calculate cosine similarity
        similarity = np.dot(feat1, feat2)
        
        # Different thresholds for different models
        threshold = 0.25 if self.model_name == "LightCNN_29Layers_V2" else 0.5
        is_same = similarity > threshold

        return similarity, is_same

# Global model cache
_models = {}

def get_model(model_name: ModelType = "LightCNN-V4", device: str = None) -> LightCNN:
    global _models
    key = f"{model_name}_{device}"
    if key not in _models:
        _models[key] = LightCNN(model_name=model_name, device=device)
    return _models[key]

def verify(
    image1: Union[str, np.ndarray], 
    image2: Union[str, np.ndarray], 
    model_name: ModelType = "LightCNN-V4",
    align: bool = True
) -> Tuple[float, bool]:
    model = get_model(model_name)
    return model.verify(image1, image2, align=align)

def get_features(
    image: Union[str, np.ndarray], 
    model_name: ModelType = "LightCNN-V4",
    align: bool = True
) -> np.ndarray:
    model = get_model(model_name)
    return model.get_features(image, align=align)
